#include <anchor_relax.hh>

#include <protocols/relax/Ramady.hh>

#include <protocols/moves/Mover.fwd.hh>
#include <protocols/moves/MinMover.hh>
#include <protocols/moves/PackRotamersMover.hh>

#include <core/pack/task/PackerTask.hh>
#include <core/pack/task/TaskFactory.hh>
#include <core/pose/Pose.hh>
#include <core/pose/util.hh>

#include <core/kinematics/MoveMap.hh>
#include <core/kinematics/util.hh>

#include <core/chemical/util.hh>
#include <core/chemical/ChemicalManager.fwd.hh>

#include <core/scoring/constraints/HarmonicFunc.hh>
#include <core/scoring/ScoreFunction.hh>
#include <core/scoring/ScoreType.hh>
#include <core/options/option.hh>
#include <core/options/keys/in.OptionKeys.gen.hh>
#include <core/options/keys/relax.OptionKeys.gen.hh>


#include <utility/string_util.hh>
#include <ObjexxFCL/string.functions.hh>


# include <mr.hh>
# include <premr_options.hh>

#include <numeric/random/random.fwd.hh>


AnchorRelax::~AnchorRelax() {
}

AnchorRelax::AnchorRelax(core::scoring::ScoreFunctionOP scorefxn, core::Size kkkk):
  RelaxProtocolBase("AnchorRelax", scorefxn),  min_type_("dfpmin_armijo_nonmonotone") { 
  // checkpoints_ ("AnchorRelax"), movemap_tag_(NULL) {
  AnchorRelax::make_relax_template(0);
  // AnchorRelax::set_and_create_mr_defaults();
}

// void AnchorRelax::set_molecular_replacement(MolecularReplacement mr) {
//   AnchorRelax::mr_ = mr;
// }
// 
// MolecularReplacement AnchorRelax::get_molecular_replacement() {
//   AnchorRelax::set_and_create_mr_defaults(); 
//   return AnchorRelax::mr_;
// }

void AnchorRelax::make_relax_template( core::Size standard_repeats ) {
  using namespace ObjexxFCL;
  using namespace std;
  templates_.clear();
  vector<std::string> lines;
  string line;

  lines.push_back("repeat " + string_of(standard_repeats));
  lines.push_back("ramp_repack_min 0.02  0.01     1.0"      );
  lines.push_back("ramp_repack_min 0.250 0.01     0.5"      );
  lines.push_back("ramp_repack_min 0.550 0.01     0.0"      );
  lines.push_back("ramp_repack_min 1     0.00001  0.0"      );
  lines.push_back("accept_to_best"                  );
  lines.push_back("endrepeat "  ); 

  int linecount = 0;
  templates_.clear();

  core::Size i;
  for( i =0; i< lines.size(); i++ ){
    line = lines[i];
    linecount++;
    std::vector< std::string > tokens ( utility::split( line ) );
  
    if ( tokens.size() > 0 ) {
      RelaxScriptCommand newcmd;
      newcmd.command = tokens[0];
      if (tokens.size() > 1) {newcmd.param1 = atof(tokens[1].c_str()); newcmd.nparams = 1;}
      if (tokens.size() > 2) {newcmd.param2 = atof(tokens[2].c_str()); newcmd.nparams = 2;}
      if (tokens.size() > 3) {newcmd.param3 = atof(tokens[3].c_str()); newcmd.nparams = 3;}
      if (tokens.size() > 4) {newcmd.param4 = atof(tokens[4].c_str()); newcmd.nparams = 4;}
      templates_.push_back( newcmd );
    }
  }

}

void AnchorRelax::apply(core::pose::Pose &pose) {
  
  bool do_rama_repair = ramady_;
  if(numeric::random::uniform() <= 0.1) do_rama_repair = false;

  if(!pose.is_fullatom()) {
    core::chemical::switch_to_residue_type_set(pose, core::chemical::FA_STANDARD);
  }

  core::kinematics::MoveMapOP local_movemap = get_movemap()->clone();
  //initiallize_movemap( pose, *local_movemap);

  set_up_constraints(pose, *local_movemap);
  core::scoring::ScoreFunctionOP local_scorefxn(get_scorefxn()->clone());
  core::scoring::EnergyMap full_weights = local_scorefxn()->weights();
 
  //we have a task 
  core::pack::task::PackerTaskOP task_;
  if ( get_task_factory() != 0 ) {
    task_ = get_task_factory()->create_task_and_apply_taskoperations( pose );
  } else {
    task_ = core::pack::task::TaskFactory::create_packer_task( pose ); //!!!!!!!!!!! << who tf put this comment here and why ?
    utility::vector1<bool> allow_repack( pose.total_residue(), repack_);
  
    if ( !core::options::option[ core::options::OptionKeys::relax::chi_move].user() ) {
      for ( Size pos = 1; pos <= pose.total_residue(); pos++ ) {
        allow_repack[ pos ] = local_movemap->get_chi( pos );
      }
    }
    task_->initialize_from_command_line().restrict_to_repacking().restrict_to_residues(allow_repack);
    task_->or_include_current( true );
  }

  protocols::moves::PackRotamersMoverOP pack_full_repack_ = 
                                        new protocols::moves::PackRotamersMover(local_scorefxn, task_);

  //side chain repacking...
  // protocols::moves::PackRotamersMoverOP pack_full_repack_;
  // pack_full_repack_->apply(pose);

  // core::Size chk_counter = 0;

  int repeat_step = 0;
  int repeat_count = -1;
  int total_repeat_count = 0;
  
  int total_count = 0;
  for ( core::Size ncmd = 0; ncmd < templates_.size(); ncmd ++ ){
    total_count++;
    RelaxScriptCommand cmd = templates_[ncmd];
    if(cmd.command == "repeat") {
      if(cmd.nparams < 1) {}
      repeat_count = (int) cmd.param1;
      repeat_step = ncmd;
    }
    if( cmd.command == "endrepeat" ){
      //TR.Debug << "CMD:  Repeat: " << repeat_count << std::endl;
      repeat_count -- ;
      total_repeat_count ++ ;
      if( repeat_count <= 0 ){}
      else{
        ncmd = repeat_step;
      }
    }
    if( cmd.command == "dump" ){
      if( cmd.nparams < 1 ){ utility_exit_with_message( "ERROR: Syntax: " + cmd.command + "<number> " ); }
      pose.dump_pdb( "dump_" + ObjexxFCL::right_string_of( (int) cmd.param1, 4, '0' ) );
    }
    if( cmd.command == "ramp_repack_min" ){
      if( cmd.nparams < 2 ){ utility_exit_with_message( "More parameters expected after : " + cmd.command  ); }
      // The first parameter is the relate repulsive weight
      local_scorefxn->set_weight( core::scoring::fa_rep, full_weights[ core::scoring::fa_rep ] * cmd.param1 );
      // The third paramter is the coordinate constraint weight
      if( constrain_coords() && (cmd.nparams >= 3) ){
         local_scorefxn->set_weight( core::scoring::coordinate_constraint,  
                                      full_weights[ core::scoring::coordinate_constraint ] * cmd.param3 );
      }
      // decide when to call ramady repair code
      if( total_repeat_count > 1 && repeat_count > 2 ){
        if( cmd.param1 < 0.2 ){
          if( do_rama_repair ){
            protocols::relax::fix_worst_bad_ramas( pose, 1, 0.0);
          }
        }
      }
      // chk_counter++;
      // std::string checkpoint_id = "chk" + string_of( chk_counter );
      // if (!checkpoints_.recover_checkpoint( pose, get_current_tag(), checkpoint_id, true, true )){
      pack_full_repack_->apply( pose );
      do_minimize_and_mr( pose, cmd.param2, local_movemap, local_scorefxn );
        // checkpoints_.checkpoint( pose, get_current_tag(), checkpoint_id,  true );
      // }
    }     
  }
  cout << "After relax score: " << (*local_scorefxn)(pose) <<endl;
}

void AnchorRelax::do_minimize_and_mr(core::pose::Pose &pose, core::Real tolerance, 
                                     core::kinematics::MoveMapOP local_movemap, 
                                     core::scoring::ScoreFunctionOP local_scorefxn) {
  using namespace core::scoring;
  using namespace core::conformation;


  protocols::moves::MinMoverOP min_mover;
  min_mover = new protocols::moves::MinMover(local_movemap, local_scorefxn, min_type_, tolerance, true);
  min_mover->apply(pose);
  // mr_.run_phaser(pose, "xxxx");
}

void AnchorRelax::relax_directly_from_rosetta(const Pose pose, FullScorefxnOP full_scorefxn) {
  // relax::FastRelax fast_relax(full_scorefxn);
  // fast_relax.apply(pose); 
}

std::string AnchorRelax::get_name() const {
  return "AnchorRelax";  
}
