// #include <utility/vector1.fwd.hh>
#include <anchor_relax.hh>
#include <protocols/loops/loops_main.hh>
#include <core/scoring/rms_util.hh>

// ObjexxFCL Headers
// #include <ObjexxFCL/formatted.o.hh>
#include <ObjexxFCL/string.functions.hh> //for pretty output.

#include <map>
#include <vector>
#include <time.h>

#include <disorder_remodel.hh>
#include <util.hh>

using namespace std;
// static numeric::random::RandomGenerator ruRG(20200);

DisorderModel::DisorderModel(string new_disorder_score_path, string new_input_model_path) {
  disorder_score_path = new_disorder_score_path;
  input_model_path    = new_input_model_path;
}

void DisorderModel::set_MR(const MolecularReplacement mr){
  mr_ = mr;
}

void DisorderModel::set_num_of_models(string new_num_of_models) {
  unsigned int num_of_models; 
  TypeCast::string2numeric(num_of_models, new_num_of_models);
  num_of_models_ = num_of_models;
}

unsigned int DisorderModel::get_num_of_models() {
  return num_of_models_;
} 

string DisorderModel::get_file_name_wo_ext() {
  return OutputFormat::get_file_name_wo_extension(input_model_path); 
}

void DisorderModel::set_native_pose(const string path2native) {
  if(path2native.empty()) return; 
  path2native_ = path2native;
  core::io::pdb::centroid_pose_from_pdb(crystal_pose_, path2native_);
}

bool DisorderModel::initialize(Rosetta n_rosetta) {
  map <int, float> rsds_to_be_built;
  if (DisorderResidues::read_disorder_scores(disorder_score_path, rsds_to_be_built)) { 
    disorder_residues_.generate_disorder_rsds(rsds_to_be_built);
    disorder_residues_.generate_each_disorder_rsds_score(rsds_to_be_built);
    disorder_residues_.set_max_min_disorder_score(rsds_to_be_built); 
    // disorder_residues_.show(); 
  }
  else {
    return false;
  }

  Pose pose;
  core::io::pdb::centroid_pose_from_pdb(pose, input_model_path);
  model_evaluation_.set_defaults(pose);
  rosetta = n_rosetta;
  curr_score_fxn_ =  rosetta.get_cen_score_fxn();
  RMover r_temp   = RMover( rosetta.get_three_residue_fragments(),  
                          rosetta.get_nine_residue_fragments(), 
                          curr_score_fxn_);
  r_mover_    = r_temp; 
  start_pose_ = pose;
  return true;
}

float DisorderModel::calculate_temperature(float cur_rsd_disorder_score) {
  const float max_tem = 4.0, min_tem = 1.0;
  const float max_disorder_score = disorder_residues_.get_max_disorder_score(); 
  const float min_disorder_score = disorder_residues_.get_min_disorder_score(); 
  if(max_disorder_score > 100.00 || min_disorder_score > 100.00) {
    cout<<"Warning: input model is highly deviated."<<endl;
  }
  if(max_disorder_score <= 0.00 || min_disorder_score <= 0.00) {
    cout<<"Warning: disorder scores are not precisely calculated."<<endl;
  }
  float scale_factor = (max_tem - min_tem) / (max_disorder_score - min_disorder_score);
  return min_tem + (cur_rsd_disorder_score - min_disorder_score ) * scale_factor;

}

void DisorderModel::remodel_disorder_rsds(int job_id, float &phaser_score, char* &pose_id) {
  Pose pose = start_pose_;
  Pose lowest_score_pose = start_pose_;
  const int max_iter = 5000; 
  list<DisorderResidue> dis_rsds = disorder_residues_.get_all_disorder_rsds(); 
  list<DisorderResidue>::iterator it; 

  time_t start_time = time(NULL);
  core::Real initial_energy = (*curr_score_fxn_)(start_pose_);
  core::Real lowest_energy = initial_energy; 
  core::Real final_energy = 0.0;
  for(int i = 0; i < max_iter; i++) {
    DisorderResidue disorder_rsd = disorder_residues_.choose_rsds_to_remodel(dis_rsds);
    int cur_rsd_pos           = disorder_rsd.get_start_rsd_pos();
    float cur_dis_order_score = disorder_residues_.get_disorder_score(cur_rsd_pos);
    float cur_temperature     = calculate_temperature(cur_dis_order_score);
    if(r_mover_.apply(pose, cur_rsd_pos, cur_temperature)) {
      final_energy = (*curr_score_fxn_)(pose);
      if(final_energy < lowest_energy) {
        lowest_score_pose = pose;
        lowest_energy = final_energy;
      }
    }
  }

  //Lowest energy models
  const string pose_name = OutputFormat::get_output_format("", '0', 4, input_model_path, job_id);
  
  // if(!path2native_.empty()) {
  //   cout<<"MOR:: LE_" + pose_name + " CA_RMSD with input structure = "
  //       <<core::scoring::CA_rmsd(start_pose_, lowest_score_pose)<<" Ang."<<endl; 
  //   cout<<"MOR:: LE_" + pose_name + " CA_RMSD with native structure: "
  //       <<core::scoring::CA_rmsd(crystal_pose_, lowest_score_pose)<<" Ang."<<endl; 
  // }
  final_energy = (*curr_score_fxn_)(lowest_score_pose);
  // cout<<"MOR:: LE_" + pose_name + " coarse energy: "<<final_energy<<endl; 
  // cout<<"MOR:: LE_" + pose_name + " has difference: "<<final_energy - initial_energy<<endl; 
  // lowest_score_pose.dump_pdb("LE_" + pose_name + ".pdb");

  protocols::relax::FastRelax fast_relax(rosetta.get_fullatom_score_fxn());
  fast_relax.apply(lowest_score_pose); 
  // if(!path2native_.empty()) {
  //   cout<<"MOR:: FE_" + pose_name + " CA_RMSD with native structure: "
  //       <<core::scoring::CA_rmsd(crystal_pose_, lowest_score_pose)<<" ang."<<endl; 
  // }
  const string full_atom_pose_name = "FE_" + pose_name;
  final_energy = (*rosetta.get_fullatom_score_fxn())(lowest_score_pose);
  // cout<<"MOR:: " + full_atom_pose_name + " full energy: "<<final_energy<<endl; 
  // lowest_score_pose.dump_scored_pdb("FE_" + pose_name + ".pdb", 
  //                                  *(rosetta.get_fullatom_score_fxn()), "1");

  phaser_score = mr_.run_phaser(lowest_score_pose, full_atom_pose_name);
  pose_id = (char*)full_atom_pose_name.c_str();
  // cout<<"MOR:: pose id " <<pose_id<<endl;

  time_t finish_time = time(NULL);
  cout<<"Total elapsed time "<< pose_name <<" is " <<finish_time - start_time << " seconds."<<endl;
}

void DisorderModel::remodel_disorder_rsds(int job_id) {
  Pose pose = start_pose_;
  Pose lowest_score_pose = start_pose_;
  int max_iter = 5000; 
  list<DisorderResidue> dis_rsds = disorder_residues_.get_all_disorder_rsds(); 
  list<DisorderResidue>::iterator it; 

  time_t start_time = time(NULL);
  core::Real initial_energy = (*curr_score_fxn_)(start_pose_);
  core::Real lowest_energy = initial_energy; 
  core::Real final_energy = 0.0;
  for(int i = 0; i < max_iter; i++) {
    DisorderResidue disorder_rsd = disorder_residues_.choose_rsds_to_remodel(dis_rsds);
    int cur_rsd_pos           = disorder_rsd.get_start_rsd_pos();
    float cur_dis_order_score =  disorder_residues_.get_disorder_score(cur_rsd_pos);
    float cur_temperature     = calculate_temperature(cur_dis_order_score);
    // cout<<"Residue position "<<disorder_rsd.get_start_rsd_pos()
    //     <<" has score " <<cur_dis_order_score<<" and chosen temperature " 
    //     <<cur_temperature<<endl; 
    if(r_mover_.apply(pose, cur_rsd_pos, cur_temperature)) {
      final_energy = (*curr_score_fxn_)(pose);
      if(final_energy < lowest_energy) {
        lowest_score_pose = pose;
        lowest_energy = final_energy;
      }
    }
  }

  //Lowest energy models
  const string pose_name = OutputFormat::get_output_format("", '0', 4, input_model_path, job_id);
  if(!path2native_.empty()) {
    cout<<"MOR:: LE_" + pose_name + " CA_RMSD with input structure = "
        <<core::scoring::CA_rmsd(start_pose_, lowest_score_pose)<<" Ang."<<endl; 
    cout<<"MOR:: LE_" + pose_name + " CA_RMSD with native structure: "
        <<core::scoring::CA_rmsd(crystal_pose_, lowest_score_pose)<<" Ang."<<endl; 
  }
  final_energy = (*curr_score_fxn_)(lowest_score_pose);
  cout<<"MOR:: LE_" + pose_name + " coarse energy: "<<final_energy<<endl; 
  cout<<"MOR:: LE_" + pose_name + " has difference: "<<final_energy - initial_energy<<endl; 
  lowest_score_pose.dump_pdb("LE_" + pose_name + ".pdb");

  protocols::relax::FastRelax fast_relax(rosetta.get_fullatom_score_fxn());
  fast_relax.apply(lowest_score_pose); 
  if(!path2native_.empty()) {
    cout<<"MOR:: FE_" + pose_name + " CA_RMSD with native structure: "
        <<core::scoring::CA_rmsd(crystal_pose_, lowest_score_pose)<<" Ang."<<endl; 
  }
  final_energy = (*rosetta.get_fullatom_score_fxn())(lowest_score_pose);
  cout<<"MOR:: FE_" + pose_name + " full energy: "<<final_energy<<endl; 
  lowest_score_pose.dump_scored_pdb("FE_" + pose_name + ".pdb", *(rosetta.get_fullatom_score_fxn()), "1");

  time_t finish_time = time(NULL);
  cout<<"Total elapsed time "<< pose_name <<" is " <<finish_time - start_time << " seconds."<<endl;

}

//@brief return secondary structure element
string DisorderModel::check_ss( string const sec_ss, int pos) {
  return sec_ss.substr(pos - 1, 1);
}

