#include <protocols/loops/Loop.hh>
#include <core/options/keys/loops.OptionKeys.gen.hh>
#include <core/types.hh>

#include<list>
#include<algorithm>

#include <loop_wrapper.hh>

//// Unit Headers
#include <protocols/loops/util.hh>
#include <protocols/loops/loops_main.hh>
#include <protocols/loops/LoopMover.hh>
#include <protocols/loops/LoopMover_QuickCCD.hh>
#include <protocols/loops/Loops.hh>
#include <protocols/moves/MonteCarlo.hh>
#include <core/conformation/Residue.hh>
//// Rosetta Headers
#include <core/chemical/VariantType.hh>
#include <core/chemical/util.hh>
#include <core/conformation/symmetry/SymmetricConformation.hh>
#include <core/id/TorsionID.hh>
#include <core/kinematics/FoldTree.hh>
#include <core/kinematics/MoveMap.hh>
#include <core/optimization/AtomTreeMinimizer.hh>
#include <core/optimization/MinimizerOptions.hh>
#include <core/options/option.hh>
#include <core/pose/Pose.hh>
#include <core/scoring/Energies.hh>
#include <core/scoring/ScoreFunction.hh>
#include <core/fragment/FragSet.hh>
#include <protocols/abinitio/FragmentMover.hh>

#include <core/conformation/symmetry/util.hh>
#include <core/optimization/symmetry/SymAtomTreeMinimizer.hh>

#include <core/scoring/constraints/ConstraintSet.hh>
#include <core/conformation/symmetry/SymmetryInfo.hh>

#include <core/util/Tracer.hh> // tracer output
#include <protocols/loops/ccd_closure.hh>

//Utility Headers
#include <numeric/random/random.hh>




RosettaLoopWrapper::RosettaLoopWrapper() {}

RosettaLoopWrapper::RosettaLoopWrapper(core::scoring::ScoreFunctionOP scorefxn) {
  RosettaLoopWrapper::cur_scorefxn_ = scorefxn;
} 

RosettaLoopWrapper::~RosettaLoopWrapper() {}

void RosettaLoopWrapper::setup_loop(Size start_rsd_pos, Size end_rsd_pos) {
  float skip_rate = 0.0;
  int cut_pt (0);
  bool extend_loop = false;
  if(start_rsd_pos == end_rsd_pos) {
    if(start_rsd_pos == 1)
      end_rsd_pos++;
    else if (end_rsd_pos== 128)
      start_rsd_pos--;
    else if(start_rsd_pos == end_rsd_pos) {
      start_rsd_pos--; end_rsd_pos++;
    }
  }
  loops_.push_back( protocols::loops::Loop(start_rsd_pos , end_rsd_pos, cut_pt, skip_rate, extend_loop));
  std::cout << "Loop setup " << start_rsd_pos << " " <<end_rsd_pos << std::endl;
}

void RosettaLoopWrapper::apply_rebuild_loop(core::pose::Pose& pose, Size start_res_pos, Size end_res_pos) {
  setup_loop(start_res_pos, end_res_pos);
  protocols::LoopRebuild loop_rebuild(cur_scorefxn_, loops_); 
  loop_rebuild.apply(pose);
}


CCD_AnchorMover::CCD_AnchorMover(protocols::loops::Loops loops_in) : IndependentLoopMover(loops_in) {
  scorefxn_ = protocols::loops::get_cen_scorefxn();
  protocols::moves::Mover::type("CCD_AnchorMover");
  set_default_settings();
}

CCD_AnchorMover::CCD_AnchorMover(protocols::loops::Loops loops_in,  
  core::scoring::ScoreFunctionOP scorefxn) : IndependentLoopMover( loops_in ) {
  if ( scorefxn ) {
    scorefxn_ = scorefxn;
  } else {
    scorefxn_ = protocols::loops::get_cen_scorefxn();
  }
  protocols::moves::Mover::type("CCD_AnchorMover");
  set_default_settings();
}

CCD_AnchorMover::CCD_AnchorMover(protocols::loops::Loops loops_in, core::scoring::ScoreFunctionOP scorefxn,
  core::fragment::FragSetOP fragset) : IndependentLoopMover( loops_in) {
  if ( scorefxn ) {
    scorefxn_ = scorefxn;
  } else {
    scorefxn_ = protocols::loops::get_cen_scorefxn();
  }
  protocols::moves::Mover::type("CCD_AnchorMover");
  set_default_settings();
  add_fragments( fragset );
}

//destructors
CCD_AnchorMover::~CCD_AnchorMover(){}

protocols::moves::MoverOP CCD_AnchorMover::clone() const {
  return new CCD_AnchorMover(*this);
}

std::string CCD_AnchorMover::get_name() const {
  return "CCD_AnchorMover";
}


protocols::loops::LoopResult CCD_AnchorMover::model_loop(core::pose::Pose &pose, protocols::loops::Loop const & loop) {
  return protocols::loops::Success; 
}


void CCD_AnchorMover::build_loop(core::pose::Pose &pose, 
      protocols::loops::Loop const &loop, DisorderResidues disorder_rsds) {
  using namespace protocols::loops;
  // Loops selected_loops = loop;
  // select_loops(selected_loops);
  bool result;

  // for(Loops::iterator it = selected_loops.v_begin(), it_end = selected_loops.v_end(); it != it_end; ++it) {
  Loop buildloop = loop;
  buildloop.auto_choose_cutpoint(pose);
  result = remodel_using_CCD(pose, buildloop, disorder_rsds );
  cout <<"ROS:: success and failure " << result <<endl;
  //}
}

string CCD_AnchorMover::int2string(int number) {
  stringstream ss;     //create a stringstream
  ss << number;        //add number to the stream
  return ss.str();     //return a string with the contents of the stream
}


bool CCD_AnchorMover::remodel_using_CCD( core::pose::Pose &pose, 
                        protocols::loops::Loop const & loop, DisorderResidues disorder_rsds ) {
  using namespace core;
  using namespace kinematics;
  using namespace scoring;
  using namespace optimization;
  using namespace protocols::abinitio;
  using namespace options;
  using namespace core::options::OptionKeys;
  using namespace numeric::random;
  
  // tr.Info << "***** CCD CLOSURE *****" << std::endl; // bool debug = option[ core::options::OptionKeys::loops::debug ](); 
  core::Size const nres =  pose.total_residue();
  
  // store starting fold tree and cut pose
  kinematics::FoldTree f_orig=pose.fold_tree();
  set_single_loop_fold_tree( pose, loop );
  
  // generate movemap after fold_tree is set
  kinematics::MoveMapOP mm_one_loop = new kinematics::MoveMap(), mm_one_loop_symm = new kinematics::MoveMap();
  set_move_map_for_centroid_loop( loop, *mm_one_loop );
  set_move_map_for_centroid_loop( loop, *mm_one_loop_symm );
  if ( core::conformation::symmetry::is_symmetric( pose ) )  {
    core::conformation::symmetry::make_symmetric_movemap( pose, *mm_one_loop_symm );
  }
  
  // int const loop_size( loop.stop() - loop.start() + 1 );
  
  int cycles2 = 1; // 2; //ROS::commented
  // Minimum of 15 cycles
  // Maximum of 5*15 cycles;
  // int base_cycles( std::max( 15, static_cast<int>( 5*std::min(loop_size,15) ))); // ROS::commented 
  int base_cycles(1); 
  int cycles3 = base_cycles;
  // tr.Info << "Number of cycles: cycles2 and cycles3 " //ROS::commented...
  //   << cycles2 << " " << cycles3
  //   << std::endl;
  
  //bool chainbreak_present =  ( loop.start() != 1 && loop.stop() != nres );
  // special case ... vrt res at last position
  //chainbreak_present &= (loop.stop() != nres-1 || pose.residue( nres ).aa() != core::chemical::aa_vrt );
  bool chainbreak_present = ( loop.start() != 1 && loop.stop() != nres &&
                              !pose.residue( loop.start() ).is_lower_terminus() &&
                              !pose.residue( loop.stop() ).is_upper_terminus() );
  // set loop.cut() variant residue for chainbreak score if chanbreak is present
  if ( chainbreak_present ) {
    chemical::add_variant_type_to_pose_residue(pose, chemical::CUTPOINT_LOWER, loop.cut());
    chemical::add_variant_type_to_pose_residue(pose, chemical::CUTPOINT_UPPER, loop.cut()+1);
  }
  
  (*scorefxn_)(pose);
  core::pose::Pose start_pose = pose;
  
  // either extend or at least idealize the loop (just in case).
  if ( loop.is_extended() ) set_extended_torsions( pose, loop );
  else  idealize_loop( pose, loop );
  
  // omega needs to be moveable for FragmentMovers, so use a separate movemap
  kinematics::MoveMapOP frag_mover_movemap = new kinematics::MoveMap();
  frag_mover_movemap->set_bb_true_range( loop.start(), loop.stop() );
  //fpd need to leave movemap for fragment insertion nonsymmetric
  //    (so that insertions will happen when scoring subunit is not 1st)
  
  utility::vector1< FragmentMoverOP > fragmover;
  for ( utility::vector1< core::fragment::FragSetOP >::const_iterator
        it = frag_libs_.begin(), it_end = frag_libs_.end();
        it != it_end; it++ ) {
    ClassicFragmentMover *cfm = new ClassicFragmentMover( *it, frag_mover_movemap );
    cfm->set_check_ss(false);
    cfm->enable_end_bias_check( false );
    fragmover.push_back( cfm );
  }
  
  core::Real m_Temperature_ = 2.0;
  protocols::moves::MonteCarloOP mc_ = new protocols::moves::MonteCarlo( pose, *scorefxn_, m_Temperature_ );
 
  // int count = 0; //ROS::just for debug
  cout << "ROS:: before fragment replacement " << (*scorefxn_)(pose)<<endl; //ROS::

  //core::Real a = (*scorefxn_)(pose);
  // ROS::I will change here...
  // insert random fragment as many times as the loop is long (not quite
  // the exact same as the old code)
  // int start_rsd = 0, end_rsd = 0;
  // start_rsd = loop.start();
  // end_rsd   = loop.stop();
  // list<DisorderResidue> selected_disorder_rsds_ = disorder_rsds.get_disorder_rsds(start_rsd, end_rsd);
  // for ( Size i = loop.start(); i <= loop.stop(); ++i ) {
  //   for ( utility::vector1< FragmentMoverOP >::const_iterator
  //       it = fragmover.begin(),it_end = fragmover.end(); it != it_end; it++ ) {
  //     // DisorderResidue selected_rsd = disorder_rsds.choose_rsds_to_remodel(selected_disorder_rsds_);
  //     // (*it)->apply( pose, selected_rsd.get_start_rsd_pos());
  //     (*it)->apply( pose );
  //     //mc_->boltzmann( pose );
  //     //count++; //ROS::
  //   }
  // }
  // mc_->recover_low(pose);
  // cout << "ROS:: after fragment replacement " << (*scorefxn_)(pose)<<endl; //ROS::

  // pose = start_pose; //ROS:: we don't consider
  // cout << "ROS:: before random fragment replacement " << (*scorefxn_)(pose)<<endl; //ROS::

  if ( randomize_loop_ ) {
    // insert random fragment as many times as the loop is long (not quite
    // the exact same as the old code)
    for ( Size i = loop.start(); i <= loop.stop(); ++i ) {
      for ( utility::vector1< FragmentMoverOP >::const_iterator
          it = fragmover.begin(),it_end = fragmover.end(); it != it_end; it++
      ) {
        (*it)->apply( pose );
      }
    }
  }
  mc_->recover_low(pose);
  cout << "ROS:: after random fragment replacement " << (*scorefxn_)(pose)<<endl; //ROS::
 

  // Set up Minimizer object
  AtomTreeMinimizerOP mzr;
  float const dummy_tol( 0.001 ); // linmin sets tol internally
  bool const use_nblist( true ), deriv_check( false ); // true ); // false );
  MinimizerOptions options( "linmin", dummy_tol, use_nblist, deriv_check);
  if ( core::conformation::symmetry::is_symmetric( pose ) ) {
    mzr = new core::optimization::symmetry::SymAtomTreeMinimizer;
  } else {
    mzr = new core::optimization::AtomTreeMinimizer;
  }
  
  
  // Set up MonteCarlo Object
  core::Real const init_temp = 2.0;
  core::Real temperature = init_temp;
  mc_->reset( pose );
  mc_->set_temperature( temperature );
  
  // int   starttime    = time(NULL);
  int   frag_count   = 0;
  // scorefxn_->show_line_headers( tr.Info ); //ROS::Commented..
  // tr.Info << std::endl;
  
  core::Real const final_temp( 1.0 );
  core::Real const gamma = std::pow(
    (final_temp/init_temp), (1.0/(50)) //cycles2*cycles3))  // ROS::don't know why commented
  );
  
  bool  has_constraints         = pose.constraint_set()->has_constraints();
  float final_constraint_weight = 0.2; // option[ core::options::OptionKeys::constraints::cst_weight ](); // this is really stupid.//ROS:: put 0.2
  
  if ( chainbreak_present ) scorefxn_->set_weight( chainbreak, 1.0 );
  
  // for symmetric case, upweight the chainbreak by # of monomers
  // if ( core::conformation::symmetry::is_symmetric( pose ) ) {
  //   core::conformation::symmetry::SymmetricConformation const & symm_conf (
  //         dynamic_cast<core::conformation::symmetry::SymmetricConformation const & > ( pose.conformation()) );
  //   core::conformation::symmetry::SymmetryInfoCOP symm_info( symm_conf.Symmetry_Info() );
  //   scorefxn_->set_weight( chainbreak, symm_info->subunits() );
  // }
  
  for( int c2 = 1; c2 <= cycles2; ++c2 ) {
    mc_->recover_low( pose );
  
    // ramp up constraints
    if ( has_constraints ) {
      if( c2 != cycles2 ) {
        scorefxn_->set_weight( core::scoring::atom_pair_constraint, final_constraint_weight*float(c2)/float( cycles2 ) );
      } else {
        scorefxn_->set_weight( core::scoring::atom_pair_constraint, final_constraint_weight * 0.2 );
      }
    }
  
    (*scorefxn_)(pose);
    // if ( tr.visible() ) { scorefxn_->show_line( tr.Info , pose ); }
    // tr.Info << std::endl;
    mc_->score_function( *scorefxn_ );
  
    for ( int c3 = 1; c3 <= cycles3; ++c3 ) {
      temperature *= gamma;
      mc_->set_temperature( temperature );
      for ( utility::vector1< FragmentMoverOP >::const_iterator
          it = fragmover.begin(),it_end = fragmover.end();
          it != it_end; it++
      ) {
        (*it)->apply( pose );
        if ( chainbreak_present ) {
          fast_ccd_close_loops( pose, loop,  *mm_one_loop );
        }
  
        mzr->run( pose, *mm_one_loop_symm, *scorefxn_, options );
        mc_->boltzmann( pose );
        frag_count++;
      }
      // ROS:: I commented 
      // if ( debug ) {
      //   (*scorefxn_)(pose);
      //   scorefxn_->show_line( tr.Info , pose );
      //   tr.Info << std::endl;
      // }
    } // for c3 in cycles3
  } // for c2 in cycles2
  
  // int looptime = time(NULL) - starttime;
  // tr << "FragCount: " << frag_count << std::endl;
  // tr << "Looptime " << looptime << std::endl;
  
  pose = mc_->lowest_score_pose();
  // scorefxn_->show(  tr.Info , pose );
  // tr.Info << "-------------------------" << std::endl;
  mc_->show_counters();
  
  // restore original CB wt if symmetric
  if ( core::conformation::symmetry::is_symmetric( pose ) ) {
    scorefxn_->set_weight( chainbreak, 1.0 );
  }
  
  // Check chain break !
  if ( chainbreak_present ) {
    using namespace core::scoring;
    (*scorefxn_)( pose );
    core::Real chain_break_score = std::max(
      static_cast<float> (pose.energies().total_energies()[ chainbreak ]),
      static_cast<float> (pose.energies().total_energies()[ linear_chainbreak ])
    );
  
    core::Real chain_break_tol
      = option[ core::options::OptionKeys::loops::chain_break_tol ]();
    // tr.Info << "Chainbreak: " << chain_break_score << " Max: "
    //   << chain_break_tol << std::endl;
  
    if ( chain_break_score > (chain_break_tol*10) ) 
      return true; //ExtendFailure;   // if we have a really bad chain break, extend loop definitions
    if ( chain_break_score > chain_break_tol )      
      return false; //return Failure;         // if we only have a slight chainbreak problem, try again
  } // if ( chainbreak_present )
  
  // return to original fold tree
  protocols::loops::remove_cutpoint_variants( pose );
  pose.fold_tree( f_orig );
  return true; //Success;
}

void fast_ccd_close_loops( core::pose::Pose & pose,
  protocols::loops::Loop const & loop, core::kinematics::MoveMap & mm) {
  // param for ccd_closure
  using namespace core;
  int  const ccd_cycles = { 100 }; // num of cycles of ccd_moves
  Real const ccd_tol = { 0.01 }; // criterion for a closed loop
  bool const rama_check = { false };
  Real const max_rama_score_increase = { 2.0 }; // dummy number when rama_check is false
  Real const max_total_delta_helix = { 10.0 }; // max overall angle changes for a helical residue
  Real const max_total_delta_strand = { 50.0 }; // ... for a residue in strand
  Real const max_total_delta_loop = { 75.0 }; // ... for a residue in loop
  // output for ccd_closure
  core::Real forward_deviation, backward_deviation; // actually loop closure msd, both dirs
  core::Real torsion_delta, rama_delta; // actually torsion and rama score changes, averaged by loop_size
  protocols::loops::fast_ccd_loop_closure(
    pose, mm, loop.start() , loop.stop(), loop.cut(), ccd_cycles,
    ccd_tol, rama_check, max_rama_score_increase, max_total_delta_helix,
    max_total_delta_strand, max_total_delta_loop, forward_deviation,
    backward_deviation, torsion_delta, rama_delta
  );

  // fix secondary structure??
  for (core::Size i=loop.start(); i<=loop.stop(); ++i) {
    char ss_i = pose.conformation().secstruct( i );
    if ( ss_i != 'L' && ss_i != 'H' && ss_i != 'E')
      pose.set_secstruct( i , 'L' );
  }
}

