#ifndef MPI_JOB_DISTRIBUTION_HH
#define MPI_JOB_DISTRIBUTION_HH

#ifdef USEMPI
#include<fstream>
#include<iostream>
#include<mpi.h>
#include<disorder_remodel.hh>
#endif

#ifdef USEMPI

using namespace std;

namespace MPI_Job_Dist {

  void dump_model_tfz(ofstream &output_stream, const char* pose_id, const float tfz_score) { 
    output_stream<<pose_id;  
    output_stream.setf(ios::fixed, ios::floatfield);
    output_stream.precision(1);
    output_stream.width(6);
    output_stream<<tfz_score;
    output_stream<<endl;
  }

  void master_node_jobs(int total_cpus, const int total_jobs) {
    int sent_job = 0, finished_job=0;
    MPI_Status status;
  
    for(int rank = 1; rank < total_cpus; rank++) {
      sent_job++;
      if(sent_job > total_jobs) break;
      MPI_Send(&sent_job, 1, MPI_INT, rank, 1, MPI_COMM_WORLD);
      //cout<<"Sending jobs to node "<<rank<<" with job number "<<sent_job<<endl;
    }
    while(1) {
      MPI_Recv(&finished_job, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      cout<<"Recieve job from node "<<status.MPI_SOURCE<<" with job number "<<finished_job<<endl;
      sent_job++;
      if(sent_job > total_jobs && finished_job >= total_jobs) break;
      if(sent_job <= total_jobs) 
        MPI_Send(&sent_job, 1, MPI_INT,status.MPI_SOURCE, 1, MPI_COMM_WORLD);
    }
  
    //send message to stop....
    for (int rank = 1; rank < total_cpus; ++rank) {
      cout<<"Shutting down nodes "<<rank<<endl;
      MPI_Send(0, 0, MPI_INT, rank, 2, MPI_COMM_WORLD);
    }
  }

  void master_node_jobs_phaser( int total_cpus, const int total_jobs, 
                                const int succeed_jobs, const float min_cutoff) {
    int sent_job = 0, finished_job = 0, success_count = 0;
    float tfz_score = 0.0;
    char pose_name[19];
    MPI_Status status;
    ofstream output_file;
    output_file.open("phaser_scores.out");
    output_file<<"Template pdb   Phaser-score"<<endl;
  
    for(int rank = 1; rank < total_cpus; rank++) {
      sent_job++;
      if(sent_job > total_jobs) break;
      MPI_Send(&sent_job, 1, MPI_INT, rank, 1, MPI_COMM_WORLD);
      //cout<<"Sending jobs to node "<<rank<<" with job number "<<sent_job<<endl;
    }

    while(1) {
      MPI_Recv(&finished_job, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      cout<<"Recieve job from node "<<status.MPI_SOURCE<<" with job number "<<finished_job<<endl;
      
      //sent jobs 
      MPI_Recv(&tfz_score, 1, MPI_FLOAT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      MPI_Recv(pose_name, 19, MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      cout<<"MOR:: "<<pose_name<<" "<<tfz_score<<endl;
      if(tfz_score > min_cutoff) { 
        success_count++; 
        dump_model_tfz(output_file, pose_name, tfz_score);  
      }
      if(success_count > succeed_jobs) {
        cout<<succeed_jobs<<" solutions are found."<<endl;
        cout<<"Simulation is terminated before generating all the models."<<endl;
        break;
      }
      sent_job++;
      if(sent_job > total_jobs && finished_job >= total_jobs) break;
      if(sent_job <= total_jobs) { 
        MPI_Send(&sent_job, 1, MPI_INT,status.MPI_SOURCE, 1, MPI_COMM_WORLD);
        //cout<<"Sending jobs to node "<<status.MPI_SOURCE<<" with job number "<<sent_job<<endl;
      }
    }
    output_file.close();
    //send message to stop....
    for (int rank = 1; rank < total_cpus; ++rank) {
      cout<<"Shutting down nodes "<<rank<<endl;
      MPI_Send(0, 0, MPI_INT, rank, 2, MPI_COMM_WORLD);
    }
  }
  
  
  void slave_nodes_job(DisorderModel disorder_model) {
    MPI_Status recv_status;
    int sent_job = 0; 
    float phaser_score = 0.0;
    char* pose_name = "";
    while(1) {
      MPI_Recv(&sent_job, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &recv_status);
      if (recv_status.MPI_TAG == 2) return; 
      disorder_model.remodel_disorder_rsds(sent_job, phaser_score, pose_name); 

      MPI_Send(&sent_job, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
      MPI_Send(&phaser_score, 1, MPI_FLOAT, 0, 0, MPI_COMM_WORLD);
      MPI_Send(&pose_name[0], 19, MPI_CHAR, 0, 0, MPI_COMM_WORLD);
    }
  }
  
  void mpi_start(int argc, char* argv[], DisorderModel disorder_model) {
    int rank, total_cpus;
    unsigned int total_jobs = disorder_model.get_num_of_models();
    cout<<"# of jobs for distribution: " <<total_jobs<<endl;
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &total_cpus);
    if(rank==0) {
      master_node_jobs(total_cpus, total_jobs);
    }
    else{
      slave_nodes_job(disorder_model);
    }
    MPI_Finalize();
  }

  void mpi_start(int argc, char* argv[], DisorderModel disorder_model, 
                 const float min_threshold, const size_t num_of_succesful_models) {
    int rank, total_cpus;
    unsigned int total_jobs = disorder_model.get_num_of_models();
    cout<<"MOR:: # of jobs for distribution: " <<total_jobs<<endl;
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &total_cpus);
    if(rank==0) {
      master_node_jobs_phaser(total_cpus, total_jobs, num_of_succesful_models, min_threshold); 
    }
    else{
      slave_nodes_job(disorder_model);
    }
    MPI_Finalize();
  }
}

#endif //namespace 
#endif // MPI
