# include <mr.hh>

// # include <phaser/run/phaser_lib.h>
# include <core/pose/Pose.hh>

# include <iostream>
# include <exception>

#include<fstream>
#include<sstream>
#include<string>
#include<vector>

#include <phaser/io/PhaserError.h>
#include <phaser/io/Errors.h>
#include <phaser/io/InputMODE.h>
#include <algorithm>
#include <phaser/lib/jiffy.h>
#include <phaser/run/runPhaser.h>
#include <cctbx/sgtbx/symbols.h>
#include <cctbx/sgtbx/space_group_type.h>
#include <phaser/include/FloatType.h>


using namespace std;

MolecularReplacement::MolecularReplacement() {
  state_ = 0;
  phaser_command_ = "";
}

MolecularReplacement::MolecularReplacement(string phaser_cmd_file) {
  state_ = 1;
  if(phaser_cmd_file.empty()) 
    state_ = 0;
  else {
    string line = "";  
    ifstream command_file(phaser_cmd_file.c_str());
    if(command_file.is_open()) {
      while(command_file.good()) {
        getline(command_file, line);
        if(!line.empty()) {
          phaser_command_ += line + "\n";
        }
      }
    }
    else {
      cout<<"Warning: molecular replacement is not run."<<endl;
    }
    command_file.close();
  }
}

MolecularReplacement::MolecularReplacement(string phaser_cmd_file, string space_group, size_t num_of_copies) {
  state_ = 1;
  if(phaser_cmd_file.empty()) 
    state_ = 0;
  else {
    string line = "";  
    ifstream command_file(phaser_cmd_file.c_str());
    if(command_file.is_open()) {
      while(command_file.good()) {
        getline(command_file, line);
        if(!line.empty()) {
          phaser_command_ += line + "\n";
        }
      }
    }
    else {
      cout<<"Warning: molecular replacement is not run."<<endl;
    }
    command_file.close();
  }
  space_group_   = space_group;
  num_of_copies_ = num_of_copies;
}

void MolecularReplacement::show() {
  cout<<"MR state_ :"<<state_<<endl;
  cout<<"Phaser command "<<phaser_command_<<endl;
}

int MolecularReplacement::get_state() {
  if(state_ == 0 || state_ == 1) return state_;
  else return 0;
}

bool MolecularReplacement::call_phaser(string phaser_cmd, map<string, string> &phaser_scores) {
  phaser::Output output;
  try {
    output.setPackageCCP4();
    output.logHeader(phaser::SUMMARY,"PREPROCESSOR");
    output.logBlank(phaser::SUMMARY);
    output.logTab(0,phaser::LOGFILE,"ENTER KEYWORD INPUT FROM FILE OR FROM STANDARD INPUT");
  
    phaser::Preprocessor capture(phaser_cmd);
    cout<<capture.Echo()<<endl;
  
    phaser::phasermode PHASER_MODE;
    { //scope for input
       phaser::InputMODE inputMode(capture);
       inputMode.Analyse();
       PHASER_MODE = phaser::phasermode(inputMode.PHASER_MODE);
       if (inputMode.SUITE == phaser::CIMR) output.setPackageCIMR();
       if (inputMode.SUITE == phaser::PHENIX) output.setPackagePhenix();
       output.logKeywords(phaser::SUMMARY,capture.Echo());
       output.logTrailer();
    }
  
    switch (PHASER_MODE) {
      //auto mode separate as anisotropy included
      case phaser::MR_AUTO: { 
        phaser::af::shared<phaser::Miller> miller;
        phaser::af::shared<phaser::floatType> fobs,sigfobs;
        phaser::af::double6 uc;
        std::string f_id,sigf_id;
        std::string hall;
        { //scope for input
           output.logHeader(phaser::LOGFILE,"READ DATA FROM MTZ FILE");
           phaser::InputMR_DAT input(capture);
           input.setRESO(0,10000);
           phaser::ResultMR_DAT result = phaser::runMR_DAT(input,output); 
           output = result;
           if (result.Failure()) return false; //goto the_end_MR;
           miller = result.getMiller();
           fobs = result.getFobs();
           sigfobs = result.getSigFobs();
           uc = result.getUnitCell();
           hall = result.getHall();
           f_id = result.getLabinF();
           sigf_id = result.getLabinSIGF();
        }
        { //scope for input
           output.logHeader(phaser::LOGFILE,"AUTOMATED MOLECULAR REPLACEMENT");
           phaser::InputMR_AUTO input(capture);
           input.setREFL(miller,fobs,sigfobs);
           input.setREFL_ID(f_id,sigf_id);
           input.setCELL6(uc);
           input.setSPAC_HALL(hall);
           phaser::ResultMR result = phaser::runMR_AUTO(input,output);
           output = result;
           if (result.Failure()) return false; //goto the_end_MR;
           string sresult = result.getTopAnnotation();
           istringstream is_result(sresult);
           do {
              string sub_is_result;
              is_result >> sub_is_result;
  
              std::string item;
              std::stringstream word(sub_is_result);
  
              size_t counter = 1;
              string first_value, second_value;
              while(std::getline(word, item, '=')) {
                if(counter == 1) first_value  = item;
                if(counter == 2) second_value = item;
                counter++;
              }
              phaser_scores[first_value] = second_value;
           } while(is_result);
           return true;
        } 
      }
    }
  }
  catch (phaser::InputError& err) {
    output.logKeywords(phaser::SUMMARY,err.Echo());
    output.logTrailer(err);
  }
  catch (phaser::PhaserError& err) {
    output.logTrailer(err);
  }
  catch (std::bad_alloc const& err) {
    output.logTrailer(phaser::PhaserError(phaser::MEMORY,err.what()));
  }
  catch (std::exception const& err) {
    output.logTrailer(phaser::PhaserError(phaser::UNHANDLED,err.what()));
  }
  catch (...) {
    output.logTrailer(phaser::PhaserError(phaser::UNKNOWN));
  }
} 

float MolecularReplacement::run_phaser(Pose pose, string tag) {
  float result = 0.0; //int m = 0; int static n = 0;

  string pdbpath= "./" + tag + ".pdb";

  pose.dump_pdb(pdbpath);
  const string phaser_cmd = replace_keyword(pdbpath, tag);
  if(phaser_cmd.empty()) return 0.0; 

  map<string, string> phaser_scores;
  // phaser_scores.clear();

  cout<<"LEN:: go there"<<endl;
  
  if(!call_phaser(phaser_cmd, phaser_scores)) 
    return 0.0; 

  // cout<<"LEN:: "<<phaser_scores.size()<<endl;
  if(phaser_scores.empty())
    return 0; 
  else { 
  //   cout<<"TFZ:: "<<tag<<" "<<phaser_scores["TFZ"].c_str()<<endl;
  //   cout<<"RFZ:: "<<tag<<" "<<phaser_scores["RFZ"].c_str()<<endl;
  //   for(map<string, string>::iterator mit = phaser_scores.begin(); 
  //                                     mit != phaser_scores.end(); 
  //                                     mit++) {
  //     cout<<"MOR:: "<<tag<<" "<<mit->first<<" "<<mit->second<<endl;
  //   }
    return evaluate_phaser_score(phaser_scores);
  }
}

float MolecularReplacement::evaluate_phaser_score(map<string,string> &phaser_scores) {
  // string sg = space_group_;
  // if(!sg.empty()) {
  //   transform(sg.begin(), sg.end(), sg.begin(), ::toupper); 
  //   if(sg == "P1") {
  //     if(num_of_copies_ > 1) {
  //       if(atof(phaser_scores["TFZ"].c_str()) == 100.00)
  //         return 0.0;
  //       else 
  //         return atof(phaser_scores["TFZ"].c_str());
  //     }
  //     else
  //       return atof(phaser_scores["RFZ"].c_str());
  //   }
  //   return atof(phaser_scores["TFZ"].c_str()); 
  // }
  if(atof(phaser_scores["TFZ"].c_str()) == 100.00) 
    return atof(phaser_scores["RFZ"].c_str()); 
  else
    return atof(phaser_scores["TFZ"].c_str()); 
}
