#ifndef R_MOVER_H 
#define R_MOVER_H 

//from rosetta
#include <core/types.hh>
#include <core/pose/Pose.fwd.hh>
#include <core/kinematics/MoveMap.hh>
#include <core/fragment/Frame.fwd.hh>
#include <core/fragment/FragSet.hh>
#include <core/fragment/FragSet.fwd.hh>
#include <core/fragment/FrameList.fwd.hh>
#include <protocols/abinitio/GunnCost.hh>
#include <protocols/abinitio/FragmentMover.fwd.hh>
#include <protocols/loops/Loops.fwd.hh>
#include <protocols/moves/MonteCarlo.hh>

#include <protocols/relax/FastRelax.hh>


#include <anchor_mover.hh>
#include <loop_wrapper.hh> 

class RMover {
    
  private:   
    typedef protocols::abinitio::ClassicFragmentMoverOP ClassicFragmentMoverOP;
    typedef core::pose::Pose Pose;
    
    //move map
    core::kinematics::MoveMapOP movemap_; 

    // Large and small fragments
    protocols::abinitio::FragmentMoverOP brute_move_large_;
    protocols::abinitio::FragmentMoverOP brute_move_small_;
    protocols::abinitio::FragmentMoverOP smooth_move_small_;
    protocols::abinitio::FragmentMoverOP anchor_mover_;
    core::scoring::ScoreFunctionOP curr_score_fxn_;  //score0

    core::fragment::FragSetOP fragset_large_;
    core::fragment::FragSetOP fragset_small_;

    RosettaLoopWrapper loop_remodel_;
      
    protocols::moves::MonteCarloOP mc_; 

    // Pose lowest_score_pose_;


  public:
    
    RMover() {
    }

    RMover( core::fragment::FragSetOP fragset_small, core::fragment::FragSetOP fragset_large, 
                                                          core::scoring::ScoreFunctionOP curr_score_fxn) {
      core::kinematics::MoveMapOP movemap = new core::kinematics::MoveMap;  
      movemap->set_bb(true);
      movemap_ = movemap;
      
      ClassicFragmentMoverOP bms, bml, sms, ams;
      bms = new protocols::abinitio::ClassicFragmentMover(fragset_small, movemap);
      bml = new protocols::abinitio::ClassicFragmentMover(fragset_large, movemap); 
      sms = new protocols::abinitio::SmoothFragmentMover(fragset_small, movemap, new protocols::abinitio::GunnCost);
      ams = new AnchorMover(fragset_small, movemap, new protocols::abinitio::GunnCost);

      brute_move_small_ = bms;
      brute_move_large_ = bml;
      smooth_move_small_ = sms;
      anchor_mover_ = ams;

      curr_score_fxn_ = curr_score_fxn;
      loop_remodel_ = RosettaLoopWrapper(curr_score_fxn_); 
      
      fragset_large_ = fragset_large;
      fragset_small_ = fragset_small;
    }
    
    ~RMover() {
    }

    protocols::loops::Loops setup_loops(core::pose::Pose pose, core::Size start_rsd_pos, core::Size end_rsd_pos ) {
      protocols::loops::Loops loops_;
      loops_.push_back(setup_loop( pose, start_rsd_pos, end_rsd_pos)); 
      return loops_;
    }

    protocols::loops::Loop setup_loop(core::pose::Pose &pose, core::Size start_rsd_pos, core::Size end_rsd_pos) {
      float skip_rate = 0.0;
      int cut_pt (0);
      bool extend_loop = false;
      if(start_rsd_pos == end_rsd_pos) {
        if(start_rsd_pos == 1)
          end_rsd_pos++;
        else if (end_rsd_pos==pose.n_residue())
          start_rsd_pos--;
        else if (start_rsd_pos == end_rsd_pos) {
          start_rsd_pos--; 
          end_rsd_pos++;
        }
      }
      return protocols::loops::Loop(start_rsd_pos , end_rsd_pos, cut_pt, skip_rate, extend_loop);
    }

    void set_default_ccd() {

    }

    protocols::moves::MonteCarloOP get_mc() {
      return mc_;
    }

    // Pose get_lowest_score_pose() {
    //   return lowest_score_pose_;
    // }

    bool apply(Pose &pose, unsigned int position, float m_temperature) {
      protocols::moves::MonteCarloOP mc = new protocols::moves::MonteCarlo( pose, *curr_score_fxn_, m_temperature);
      mc_ = mc;
      // core::Real start_score, end_score; 
      // start_score = (*curr_score_fxn_)(pose);
      core::pose::Pose local_pose = pose;
      if(brute_move_small_->apply(pose, position)) {
        return mc->boltzmann(pose);
      }
      else 
        return false;
    }

    void apply(Pose & pose, DisorderResidue dis_rsd, DisorderResidues dis_rsds, float c_temperature) {
      unsigned int counter = 0, selected_rsd_pos = 1;
      unsigned int start_rsd_pos = dis_rsd.get_start_rsd_pos(); 
      unsigned int end_rsd_pos = dis_rsd.get_end_rsd_pos(); 
      if(start_rsd_pos > end_rsd_pos) return;
      const unsigned int num_of_times = 10 * (end_rsd_pos - start_rsd_pos); 

      list<DisorderResidue> cur_dis_rsds = dis_rsds.get_disorder_rsds(start_rsd_pos, end_rsd_pos);
      float total_dis_score = dis_rsds.get_total_disorder_score(cur_dis_rsds);
      cout <<"Local disorder scores "<<endl;
      while(++counter <= num_of_times) {
        selected_rsd_pos = dis_rsds.choose_local_rsds_to_remodel(cur_dis_rsds, total_dis_score);
        apply(pose, selected_rsd_pos, c_temperature);
      }
    }

    void set_score_fxn(core::scoring::ScoreFunctionOP curr_score_fxn) {
      curr_score_fxn_ = curr_score_fxn;
    }

    core::kinematics::MoveMapOP get_move_map() {
      return movemap_;
    }
};

#endif
